from pathlib import Path

import sh

from autograder.testcase_utils.abstract_testcase import (
    ArgList,
    ShCommand,
    TestCase as AbstractTestCase,
)
from autograder.testcase_utils.shell import Command, EMPTY_COMMAND


class TestCase(AbstractTestCase):
    source_suffix = ".c"
    executable_suffix = ".out"
    helper_module = "test_helper.c"
    SUBMISSION_COMPILATION_ARGS = ("-Dscanf_s=scanf", "-Dmain=__student_main__")
    compiler = Command("gcc")

    @classmethod
    def is_installed(cls) -> bool:
        return cls.compiler is not EMPTY_COMMAND

    @classmethod
    def get_template_dir(cls):
        return cls.type_source_file.parent / "c_templates"

    @classmethod
    def precompile_submission(
        cls,
        submission: Path,
        student_dir: Path,
        possible_source_file_stems: str,
        arglist,
    ):
        """Links student submission without compiling it.
        It is done to speed up total compilation time
        """
        copied_submission = super().precompile_submission(
            submission,
            student_dir,
            [submission.stem],
            arglist,
        )
        precompiled_submission = copied_submission.with_suffix(".o")
        try:
            cls.compiler(
                "-c",
                f"{copied_submission}",
                "-o",
                precompiled_submission,
                *cls.SUBMISSION_COMPILATION_ARGS,
                *arglist,
            )
        finally:
            copied_submission.unlink()
        return precompiled_submission

    def precompile_testcase(self):
        self.compiler(
            "-c",
            self.path,
            "-o",
            self.path.with_suffix(".o"),
            *self.argument_lists[ArgList.TESTCASE_PRECOMPILATION],
        )
        self.path.unlink()
        self.path = self.path.with_suffix(".o")

    def compile_testcase(self, precompiled_submission: Path) -> ShCommand:
        executable_path = self.make_executable_path(precompiled_submission)
        self.compiler(
            "-o",
            executable_path,
            precompiled_submission.with_name(self.path.name),
            str(precompiled_submission),
            *self.argument_lists[ArgList.TESTCASE_COMPILATION],
        )
        return sh.Command(executable_path)



from contextlib import contextmanager
from io import StringIO
from pathlib import Path


class TestCaseIO:
    """Represents everything related to stdin/stdout in testcase_utils"""

    def __init__(self, testcase_name_stem, formatters, input_dir, output_dir):
        self.name = testcase_name_stem
        self.formatter = formatters.get(self.name, None) or formatters.get("ALL", None)
        self.output_file: Path = output_dir / f"{self.name}.txt"

        if self.output_file.exists() and self.output_file.is_file():
            with self.output_file.open() as f:
                self.expected_output_backup = f.read()
                self.expected_output = self.format_output(self.expected_output_backup)
            with self.output_file.open("w") as f:
                f.write("")
        else:
            self.expected_output_backup = ""
            self.expected_output = ""

        input_file: Path = input_dir / f"{self.name}.txt"
        if input_file.exists() and input_file.is_file():
            with input_file.open() as f:
                self._input = StringIO(f.read().strip())
        else:
            self._input = StringIO("")

    def format_output(self, output: str) -> str:
        return output if self.formatter is None else self.formatter(output)

    @contextmanager
    def input(self):
        try:
            yield self._input
        finally:
            self._input.seek(0)

    def expected_output_equals(self, output: str) -> bool:
        return self.format_output(output) == self.expected_output

    @classmethod
    def get_empty_io(cls):
        io = TestCaseIO("/*", {}, Path(), Path())
        io.expected_output = ""
        io._input = StringIO("")
        return io

    def cleanup(self):
        if self.output_file.exists() and self.output_file.is_file():
            with self.output_file.open("w") as f:
                f.write(self.expected_output_backup)



import shutil
from typing import Protocol, Any, Optional

import sh

EMPTY_COMMAND = sh.Command("false")

class ShCommand(Protocol):
    """We use this to imitate sh.Command by duck-typing it"""

    def __call__(self, *args: str, **kwargs: Any) -> Optional[sh.RunningCommand]:
        raise NotImplementedError()


def get_stderr(error: sh.ErrorReturnCode, string):
    error_str = str(error)
    # Remove all unrelated output
    formatted_error = string + error_str[error_str.find("STDERR:") + len("STDERR") :]
    return formatted_error


def Command(command: str, *args: Any, **kwargs: Any) -> sh.Command:
    """An API for commands that postpone throwing non-existence errors from creation to runtime"""
    return EMPTY_COMMAND if shutil.which(command) is None else sh.Command(command, *args, **kwargs)



import configparser
from pathlib import Path
from typing import Any, Callable, Dict, List, TypeVar

from .testcase_utils.abstract_testcase import ArgList

DEFAULT_FILE_STEM = "Homework"


class GradingConfig:
    timeouts: Dict[str, float]
    generate_results: bool
    parallel_grading_enabled: bool
    stdout_only_grading_enabled: bool
    total_points_possible: int
    total_score_to_100_ratio: float

    assignment_name: str
    any_submission_file_name_is_allowed: bool
    possible_source_file_stems: List[str]
    testcase_weights: Dict[str, float]
    cli_argument_lists: Dict[ArgList, Dict[str, List[str]]]

    def __init__(self, config: Path, default_config: Path):
        self._configure_grading(config, default_config)

    def _configure_grading(
        self,
        config_path: Path,
        default_config_path: Path,
    ):
        cfg = self._read_config(config_path, default_config_path)

        self.timeouts = parse_config_list(cfg["TIMEOUT"], float)
        self.generate_results = cfg.getboolean("GENERATE_RESULTS")
        self.parallel_grading_enabled = cfg.getboolean("PARALLEL_GRADING_ENABLED")
        self.stdout_only_grading_enabled = cfg.getboolean("STDOUT_ONLY_GRADING_ENABLED")

        self.total_points_possible = cfg.getint("TOTAL_POINTS_POSSIBLE")
        self.total_score_to_100_ratio = self.total_points_possible / 100

        self.assignment_name = cfg["ASSIGNMENT_NAME"]

        source = cfg["POSSIBLE_SOURCE_FILE_STEMS"].strip()
        if source == "AUTO":
            source = DEFAULT_FILE_STEM
            self.any_submission_file_name_is_allowed = True
        else:
            self.any_submission_file_name_is_allowed = False
        self.possible_source_file_stems = source.replace(" ", "").split(",")

        self.testcase_weights = parse_config_list(cfg["TESTCASE_WEIGHTS"], float)

        # TODO: Name me better. The name is seriously bad
        self.cli_argument_lists = parse_arglists(cfg)

    @staticmethod
    def _read_config(path_to_user_config: Path, path_to_default_config: Path) -> configparser.SectionProxy:
        default_parser = configparser.ConfigParser()
        default_parser.read(str(path_to_default_config))

        user_parser = configparser.ConfigParser()
        user_parser.read_dict(default_parser)
        user_parser.read(str(path_to_user_config))

        return user_parser["CONFIG"]

    def generate_arglists(self, file_name: str) -> Dict[ArgList, List[str]]:
        arglist = {}
        for arglist_index, arglists_per_file in self.cli_argument_lists.items():
            if file_name in arglists_per_file:
                arglist[arglist_index] = arglists_per_file[file_name]
            elif "ALL" in arglists_per_file:
                arglist[arglist_index] = arglists_per_file["ALL"]
            else:
                arglist[arglist_index] = tuple()
        return arglist


T = TypeVar("T")


def parse_config_list(config_line: str, value_type: Callable[[Any], T]) -> Dict[str, T]:
    """Reads in a config line in the format: 'file_name:value, file_name:value, ALL:value'
    ALL sets the default value for all other entries
    """
    config_entries = {}
    for config_entry in config_line.split(","):
        if config_entry.strip():
            testcase_name, value = config_entry.strip().split(":")
            config_entries[testcase_name.strip()] = value_type(value)
    return config_entries


def parse_arglists(cfg: configparser.SectionProxy) -> Dict[ArgList, Dict[str, List[str]]]:
    argument_lists = {n: {} for n in ArgList}
    for arg_list_enum in ArgList:
        arg_list = parse_config_list(cfg[arg_list_enum.value], convert_to_arglist)
        for testcase_name, args in arg_list.items():
            argument_lists[arg_list_enum][testcase_name] = args
    return argument_lists


def convert_to_arglist(s: str) -> List[str]:
    return str(s).replace("  ", " ").strip().split()



import importlib.util
import re
import sys
from pathlib import Path


class AutograderError(Exception):
    pass


RESULT_REGEX = re.compile(r"Result: (\d+)/\d+")


def print_results(paths, min_score: int, *args, **kwargs):
    with open(paths.output_summary) as f:
        contents = f.read()
    student_outputs = contents.split("\n\n")
    for output in student_outputs:
        match = RESULT_REGEX.search(output)
        if match is not None:
            score = int(match.group(1))
            if score >= min_score:
                print(output + "\n", *args, **kwargs)


def import_from_path(module_name: str, path: Path):
    spec = importlib.util.spec_from_file_location(module_name, path)
    if spec is None or spec.loader is None:
        raise TypeError(f"File loader for {path} was not found. Please, refer to importlib docs.")
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    # TODO: Loader has no attribute exec_module. What's up with that?
    spec.loader.exec_module(module)  # type: ignore
    return module


def get_file_stems(dir_: Path):
    return (p.stem for p in dir_.iterdir()) if dir_.exists() else []
